
import datetime
import calendar
import hashlib
import logging
import os
import re
import shlex
import subprocess
import time

from tornado import gen
from tornado import ioloop

from thingosdci import settings


CONTAINER_STATE_RUNNING = 'running'
CONTAINER_STATE_EXITED = 'exited'
CONTAINER_STATE_REMOVED = 'removed'

_CONTAINER_NAME_PREFIX = 'thingosdci-{repo}-'

logger = logging.getLogger(__name__)

_containers_by_id = {}


class ContainerException(Exception):
    pass


class Container:
    def __init__(self, _id, name, created_time):
        self.id = _id
        self.name = name
        self.created_time = created_time

        self.exit_code = None

        self._removed = False
        self._state_change_callbacks = []

    def __str__(self):
        return 'container {}'.format(self.id)

    def get_age(self):
        return int(time.time() - self.created_time)

    def get_state(self):
        if self.exit_code is None:
            return CONTAINER_STATE_RUNNING

        if not self._removed:
            return CONTAINER_STATE_EXITED

        return CONTAINER_STATE_REMOVED

    def set_exited(self, exit_code):
        if self.exit_code is not None:
            raise ContainerException('container already exited')

        self.exit_code = exit_code
        logger.debug('%s has exited (exit_code=%d)', self, exit_code)

        self._run_state_change_callbacks()

    def set_removed(self):
        if self._removed:
            raise ContainerException('container already removed')

        self._removed = True
        logger.debug('%s has been removed', self)

        self._run_state_change_callbacks()

    def add_state_change_callback(self, callback):
        self._state_change_callbacks.append(callback)

    def _run_state_change_callbacks(self):
        io_loop = ioloop.IOLoop.current()
        state = self.get_state()

        for callback in self._state_change_callbacks:
            io_loop.spawn_callback(callback, state)


class DockerException(Exception):
    pass


def _cmd(cmd, pipe_stdio=True):
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)

    docker_base_cmd = shlex.split(settings.DOCKER_COMMAND)

    cmd = docker_base_cmd + cmd
    # logger.debug('executing "%s"', cmd)

    stdin = stdout = stderr = None
    if pipe_stdio:
        stdin = stdout = stderr = subprocess.PIPE

    p = subprocess.Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr, universal_newlines=True)
    stdout, stderr = p.communicate()

    if p.returncode and stderr is not None:
        raise DockerException(stderr.strip())

    return stdout


def _make_log_path(container_id):
    return os.path.join(settings.BUILD_LOGS_DIR, 'build-{}.log'.format(container_id))


def _save_log(container_id):
    log_path = _make_log_path(container_id)
    log = _cmd(['logs', container_id])

    with open(log_path, 'w') as f:
        f.write(log)

    return log


def _remove_container(container_id):
    _cmd(['container', 'rm', container_id])


def _kill_container(container_id):
    _cmd(['container', 'kill', container_id])


def _list_containers():
    name_prefix = _CONTAINER_NAME_PREFIX.format(repo=re.sub('[^a-z0-9]', '-', settings.REPO, re.IGNORECASE))
    container_info_list = []
    s = _cmd(['container', 'ls', '-a', '--no-trunc', '--format',
              '{{.ID}}%{{.Names}}%{{.CreatedAt}}%{{.Status}}'])

    lines = s.split('\n')
    lines = [l.strip() for l in lines if l.strip()]

    for line in lines:
        parts = line.split('%')

        container_id = parts[0]
        name = parts[1]
        created_at = parts[2]
        running = parts[3].startswith('Up')
        exit_code = None

        if not name.startswith(name_prefix):
            continue  # not ours

        if not running:  # determine exit code
            try:
                exit_code = int(_cmd(['wait', container_id]).strip())

            except Exception as e:
                logger.error('failed to retrieve exit code of container %s: %s', container_id, e,
                             exc_info=True)

                exit_code = 1

        created_at = ' '.join(created_at.split()[:2])
        created_at = datetime.datetime.strptime(created_at, '%Y-%m-%d %H:%M:%S')
        created_time = calendar.timegm(created_at.timetuple())

        container_info_list.append({
            'id': container_id,
            'name': name,
            'created_time': created_time,
            'running': running,
            'exit_code': exit_code
        })

    return container_info_list


@gen.coroutine
def _status_loop():
    while True:
        yield gen.sleep(1)

        try:
            container_info_list = _list_containers()

        except Exception as e:
            logger.error('failed to list docker containers: %s', e, exc_info=True)
            continue

        # see if any of the containers has exited
        seen_container_ids = set()
        for container_info in container_info_list:
            container_id = container_info['id']
            exit_code = container_info['exit_code']

            seen_container_ids.add(container_id)

            container = _containers_by_id.get(container_id)
            if not container:
                continue

            if container.get_state() == CONTAINER_STATE_RUNNING and not container_info['running']:
                container.set_exited(exit_code)

        # remove old references from _containers_by_id
        for container_id in list(_containers_by_id.keys()):
            if container_id not in seen_container_ids:
                _containers_by_id.pop(container_id)


@gen.coroutine
def _cleanup_loop():
    while True:
        yield gen.sleep(60)

        for container in _containers_by_id.values():
            # kill old containers
            state = container.get_state()
            if (state == CONTAINER_STATE_RUNNING and
                container.get_age() > settings.DOCKER_CONTAINER_MAX_AGE):

                logger.warning('%s still running after %s seconds, killing it', container, container.get_age())

                try:
                    _kill_container(container.id)

                except Exception as e:
                    logger.error('failed to kill %s: %s', container, e, exc_info=True)

            # remove the container if it's not running anymore
            elif state == CONTAINER_STATE_EXITED:
                logger.debug('saving log of %s', container)

                try:
                    _save_log(container.id)

                except Exception as e:
                    logger.error('failed to save logs of %s: %s', container, e, exc_info=True)

                logger.debug('removing %s', container)

                try:
                    _remove_container(container.id)

                except Exception as e:
                    logger.error('failed to remove %s: %s', container, e, exc_info=True)

                container.set_removed()

        # remove old logs
        for file in os.listdir(settings.BUILD_LOGS_DIR):
            path = os.path.join(settings.BUILD_LOGS_DIR, file)
            s = os.stat(path)
            age = time.time() - s.st_mtime
            if age > settings.DOCKER_LOGS_MAX_AGE:
                logger.debug('removing old log %s', path)
                try:
                    os.remove(path)

                except Exception as e:
                    logger.error('failed to remove old log %s: %s', path, e)


def run_container(env, vol, interactive=False):
    ssh_private_key_file = settings.DOCKER_COPY_SSH_PRIVATE_KEY
    if ssh_private_key_file is True:
        ssh_private_key_file = os.path.join(os.getenv('HOME'), '.ssh', 'id_rsa')

    name = _CONTAINER_NAME_PREFIX + hashlib.sha1(str(int(time.time() * 1000)).encode()).hexdigest()[:8]
    name = name.format(repo=re.sub('[^a-z0-9]', '-', settings.REPO, re.IGNORECASE))

    cmd = ['run']

    if interactive:
        cmd.append('-it')

    else:
        cmd.append('-td')

    cmd += [
        '--privileged',
        '--name', name
    ]

    if settings.DOCKER_ENV_FILE:
        cmd.append('--env-file={}'.format(settings.DOCKER_ENV_FILE))

    for k, v in env.items():
        cmd.append('-e')
        cmd.append('{}={}'.format(k, v))

    for k, v in vol.items():
        cmd.append('-v')
        cmd.append('{}:{}'.format(k, v))

    if ssh_private_key_file:
        cmd.append('-v')
        cmd.append('{}:/root/.ssh/id_rsa'.format(ssh_private_key_file))

    cmd += [
        '--cap-add=SYS_ADMIN',
        '--cap-add=MKNOD',
        settings.DOCKER_IMAGE_NAME
    ]

    if interactive:
        _cmd(cmd, pipe_stdio=not interactive)
        return None

    else:
        container_id = _cmd(cmd, pipe_stdio=not interactive).strip()

    container_info_list = _list_containers()
    for container_info in container_info_list:
        if container_info['id'] == container_id:
            container = Container(container_id, container_info['name'], container_info['created_time'])
            _containers_by_id[container_id] = container

            return container

    raise DockerException('container not present after run')


def get_container_log(container_id, last_lines=None):
    try:
        log = _save_log(container_id)

    except DockerException:
        log_path = _make_log_path(container_id)

        try:
            with open(log_path) as f:
                log = f.read()

        except Exception as e:
            logger.error('error opening log file %s: %s', log_path, e)
            log = ''

    if last_lines:
        partial_log = ''
        while last_lines > 0:
            last_lines -= 1
            p = log.rfind('\n')
            if p < 0:
                partial_log = log + partial_log
                log = ''

            else:
                partial_log = log[p:] + partial_log
                log = log[:p]

        return partial_log

    return log


def init():
    io_loop = ioloop.IOLoop.current()

    io_loop.spawn_callback(_status_loop)
    io_loop.spawn_callback(_cleanup_loop)




import logging
import os
import stat

from thingosdci import settings


_LOOP_DEV_PATTERN = '/dev/loop{}'
_FILE_PERMS = 0o660

logger = logging.getLogger(__name__)

_loop_devs = {}


class LoopDevManagerException(Exception):
    pass


def acquire_loop_dev():
    for l, busy in _loop_devs.items():
        if not busy:
            loop_dev = _LOOP_DEV_PATTERN.format(l)
            logger.debug('acquiring %s', loop_dev)
            _loop_devs[l] = True

            return loop_dev

    raise LoopDevManagerException('no free loop device')


def release_loop_dev(loop_dev):
    ld = loop_dev[9:]
    try:
        ld = int(ld)

    except ValueError:
        raise LoopDevManagerException('unknown loop device: {}'.format(loop_dev))

    try:
        busy = _loop_devs[ld]

    except KeyError:
        raise LoopDevManagerException('unknown loop device: {}'.format(loop_dev))

    if not busy:
        raise LoopDevManagerException('attempt to release free loop device: {}'.format(loop_dev))

    logger.debug('releasing %s', loop_dev)
    _loop_devs[ld] = False


def _ensure_loop_dev(loop_dev):
    if os.path.exists(loop_dev):
        return

    try:
        os.mknod(loop_dev, mode=stat.S_IFBLK | _FILE_PERMS)

    except Exception as e:
        logger.error('failed to create loop device: %s', e)


def init():
    global _loop_devs

    rng = range(settings.LOOP_DEV_RANGE[0], settings.LOOP_DEV_RANGE[1] + 1)
    logger.debug('initializing loop devices (/dev/loop%s - /dev/loop%s)', *settings.LOOP_DEV_RANGE)
    _loop_devs = {ld: False for ld in rng}

    for ld in rng:
        loop_dev = _LOOP_DEV_PATTERN.format(ld)
        _ensure_loop_dev(loop_dev)




import datetime
import functools
import logging
import os
import re

from tornado import gen
from tornado import ioloop
from tornado import web

from thingosdci import building
from thingosdci import dockerctl
from thingosdci import persist
from thingosdci import s3client
from thingosdci import settings
from thingosdci import utils
from thingosdci.reposervices import trigger


_SERVICE_CLASSES = {}
_service = None

logger = logging.getLogger(__name__)


class RepoServiceRequestHandler(web.RequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self, service):
        self.service = service

    def get(self):
        self.set_header('Content-Type', 'text/plain')
        lines = self.get_argument('lines', None)

        if lines:
            try:
                lines = int(lines)

            except ValueError:
                lines = 1

        self.finish(dockerctl.get_container_log(self.get_argument('id'), lines))


class RepoService:
    DEF_LOG_LINES = 100
    REQUEST_HANDLER_CLASS = RepoServiceRequestHandler

    def __init__(self):
        # used for nightly builds at fixed hour
        self._last_commit_by_branch = persist.load('last-commit-by-branch', {})
        self._last_nightly_commit_by_branch = persist.load('last-nightly-commit-by-branch', {})
        self._commit_ids_by_tag = persist.load('commit-ids-by-tag', {})

    def __str__(self):
        return 'reposervice'

    def make_log_url(self, build):
        return settings.WEB_BASE_URL + '/{}?id={}&lines={}'.format(self, build.container.id, self.DEF_LOG_LINES)

    def schedule_nightly_builds_for_new_commits(self):
        for branch in settings.NIGHTLY_BRANCHES:
            last_commit = self._last_commit_by_branch.get(branch)
            last_nightly_commit = self._last_nightly_commit_by_branch.get(branch)

            if last_commit and last_commit != last_nightly_commit:
                logger.debug('new commit found on branch %s', branch)
                self.schedule_nightly_build(last_commit, branch)

    def handle_pull_request_open(self, commit_id, src_repo, dst_repo, pr_no):
        logger.debug('pull request %s opened: %s -> %s (%s)', pr_no, src_repo, dst_repo, commit_id)
        if not settings.PULL_REQUESTS:
            logger.debug('pull requests ignored')
            return

        build_group = building.BuildGroup()

        for board in settings.BOARDS:
            build = building.schedule_pr_build(self, build_group, board, commit_id, pr_no)
            self._register_build(build)

        self._register_build_group(build_group)

    def handle_pull_request_update(self, commit_id, src_repo, dst_repo, pr_no):
        logger.debug('pull request %s updated: %s -> %s (%s)', pr_no, src_repo, dst_repo, commit_id)
        if not settings.PULL_REQUESTS:
            logger.debug('pull requests ignored')
            return

        build_group = building.BuildGroup()

        for board in settings.BOARDS:
            build = building.schedule_pr_build(self, build_group, board, commit_id, pr_no)
            self._register_build(build)

        self._register_build_group(build_group)

    def handle_commit(self, commit_id, branch):
        logger.debug('commit to %s (%s)', branch, commit_id)

        self._last_commit_by_branch[branch] = commit_id
        self._save_commit_info()

        if branch not in settings.NIGHTLY_BRANCHES:
            logger.debug('branch %s ignored', branch)
            return

        if settings.NIGHTLY_FIXED_HOUR is None:  # schedule build right away
            self.schedule_nightly_build(commit_id, branch)

        # else, fixed_hour_loop() will take care of it

    def handle_new_tag(self, commit_id, tag):
        if commit_id is None:
            commit_id = self._commit_ids_by_tag.get(tag)

        else:
            self._commit_ids_by_tag[tag] = commit_id
            self._save_commit_info()

        logger.debug('new tag: %s (%s)', tag, commit_id)

        if not settings.RELEASE_TAG_REGEX or not re.match(settings.RELEASE_TAG_REGEX, tag):
            logger.debug('release: tag %s ignored', tag)
            return

        version = self._prepare_version(tag)

        build_group = building.BuildGroup()

        for board in settings.BOARDS:
            build = building.schedule_tag_build(self, build_group, board, commit_id, tag, version)
            self._register_build(build)

        self._register_build_group(build_group)

    @gen.coroutine
    def handle_build_begin(self, build):
        logger.debug('handling %s begin', build)

    @gen.coroutine
    def handle_first_build_begin(self, build):
        logger.debug('handling first %s begin', build)

        if not build.group:
            return  # unlikely

        completed_builds = build.group.get_completed_builds()
        remaining_builds = build.group.get_remaining_builds()

        first_board = len(completed_builds) == 0
        if first_board:
            logger.debug('setting pending status for %s (0/%s)', build.commit_id, len(settings.BOARDS))

            yield self._set_pending(build, completed_builds, remaining_builds)
            logger.debug('status set')

    @gen.coroutine
    def handle_last_build_end(self, build):
        logger.debug('handling last %s end', build)

        completed_builds = build.group.get_completed_builds()
        failed_builds = build.group.get_failed_builds()

        if not failed_builds:
            logger.debug('setting success status for %s (%s/%s)',
                         build.commit_id, len(settings.BOARDS), len(settings.BOARDS))

            yield self._set_success(build)

            logger.debug('status set')

            group = build.group
            boards_image_files = {b.board: b.image_files for b in group.builds.values()}

            if build.type in [building.TYPE_NIGHTLY, building.TYPE_TAG]:
                yield self.handle_release(build.commit_id, build.tag, build.version,
                                          build.branch, boards_image_files, build.type)

        else:
            logger.debug('setting failed status for %s: (%s/%s)',
                         build.commit_id, len(completed_builds), len(settings.BOARDS))

            yield self._set_failed(build, failed_builds)
            logger.debug('status set')

    @gen.coroutine
    def handle_build_end(self, build):
        logger.debug('handling %s end', build)

        completed_builds = build.group.get_completed_builds()
        remaining_builds = build.group.get_remaining_builds()

        if not remaining_builds:
            return  # last build end

        logger.debug('setting pending status for %s (%s/%s)',
                     build.commit_id, len(completed_builds), len(settings.BOARDS))

        yield self._set_pending(build, completed_builds, remaining_builds)
        logger.debug('status set')

    @gen.coroutine
    def handle_release(self, commit_id, tag, version, branch, boards_image_files, build_type):
        if tag and (not settings.RELEASE_TAG_REGEX or not re.match(settings.RELEASE_TAG_REGEX, tag)):
            logger.debug('release: tag %s ignored', tag)
            return

        logger.debug('handling release on commit=%s, tag=%s, version=%s, branch=%s', commit_id, tag, version, branch)

        today = datetime.date.today()
        tag = tag or utils.branches_format(settings.NIGHTLY_TAG, branch, today)

        release = yield self.create_release(commit_id, tag, version, branch, build_type)

        for board in settings.BOARDS:
            image_files = boards_image_files.get(board)
            if not image_files:
                logger.warning('no image files supplied for board %s', board)
                continue

            for fmt in settings.IMAGE_FILE_FORMATS:
                image_file = image_files.get(fmt)
                if image_file is None:
                    logger.warning('no image files supplied for board %s, format %s', board, fmt)
                    continue

                file_name = os.path.basename(image_file)
                with open(image_file, 'rb') as f:
                    body = f.read()

                if build_type in settings.UPLOAD_SERVICE_BUILD_TYPES:
                    logger.debug('uploading image file %s (%s bytes)', image_file, len(body))
                    yield self.upload_release_file(release, board, tag, version, file_name, fmt, body)
                    logger.debug('image file %s uploaded', image_file)

                if build_type in settings.S3_UPLOAD_BUILD_TYPES:
                    logger.debug('uploading image file %s to S3 (%s bytes)', image_file, len(body))
                    s3_url = yield self._upload_release_file_s3(release, board, tag, version, file_name, fmt, body)
                    logger.debug('image file %s uploaded to S3', image_file)

                    if settings.S3_UPLOAD_ADD_RELEASE_LINK:
                        yield self.add_s3_release_link(release, board, tag, version, file_name, fmt, s3_url)

        logger.debug('release on commit=%s, tag=%s, version=%s, branch=%s completed', commit_id, tag, version, branch)

    def schedule_nightly_build(self, commit_id, branch):
        if commit_id is None:
            commit_id = self._last_commit_by_branch[branch]

        build_group = building.BuildGroup()

        for board in settings.BOARDS:
            build = building.schedule_nightly_build(self, build_group, board, commit_id, branch)
            self._register_build(build)

        self._register_build_group(build_group)

        self._last_nightly_commit_by_branch[branch] = commit_id
        self._save_commit_info()

    def _prepare_version(self, tag):
        if not settings.RELEASE_TAG_REGEX:
            return tag

        m = re.match(settings.RELEASE_TAG_REGEX, tag)
        if not m:
            return tag

        if len(m.groups()) < 2:
            return tag

        return m.group(1)

    def _register_build(self, build):
        build.add_state_change_callback(functools.partial(self._on_build_state_change, build))

    def _register_build_group(self, group):
        group.add_first_build_begin_callback(self.handle_first_build_begin)
        group.add_last_build_end_callback(self.handle_last_build_end)

    def _on_build_state_change(self, build, state):
        if state == building.STATE_RUNNING:
            self.handle_build_begin(build)

        elif state == building.STATE_ENDED:
            self.handle_build_end(build)

    def _save_commit_info(self):
        persist.save('last-commit-by-branch', self._last_commit_by_branch)
        persist.save('last-nightly-commit-by-branch', self._last_nightly_commit_by_branch)
        persist.save('commit-ids-by-tag', self._commit_ids_by_tag)

    @gen.coroutine
    def _set_pending(self, build, completed_builds, remaining_builds):
        running_remaining_builds = [b for b in remaining_builds if b.get_state() == building.STATE_RUNNING]
        if running_remaining_builds:
            running_build = running_remaining_builds[0]

        else:
            running_build = build

        url = self.make_log_url(running_build)
        description = 'building OS images ({}/{})'.format(len(completed_builds), len(settings.BOARDS))

        yield self.set_pending(build, url, description)

    @gen.coroutine
    def _set_success(self, build):
        url = self.make_log_url(build)
        description = 'OS images successfully built ({}/{})'.format(len(settings.BOARDS), len(settings.BOARDS))

        yield self.set_success(build, url, description)

    @gen.coroutine
    def _set_failed(self, build, failed_builds):
        if not failed_builds:
            logger.warning('cannot set failed status with no failed builds')
            return

        url = self.make_log_url(failed_builds[0])
        failed_boards_str = ', '.join([b.board for b in failed_builds])
        description = 'failed to build some OS images: {}'.format(failed_boards_str)

        yield self.set_failed(build, url, description)

    @gen.coroutine
    def _upload_release_file_s3(self, release, board, tag, version, name, fmt, content):
        # final URL should be in the following form:
        #    https://s3.amazonaws.com/{bucket}/{path}/{version}/{name}

        client = s3client.S3Client(access_key=settings.S3_UPLOAD_ACCESS_KEY,
                                   secret_key=settings.S3_UPLOAD_SECRET_KEY,
                                   bucket=settings.S3_UPLOAD_BUCKET)

        path = '{path}/{version}/{name}'.format(path=settings.S3_UPLOAD_PATH,
                                                version=version,
                                                name=name)

        yield client.upload(path, content)

        return 'https://s3.amazonaws.com/{bucket}/{path}/{version}/{name}'.format(bucket=settings.S3_UPLOAD_BUCKET,
                                                                                  path=settings.S3_UPLOAD_PATH,
                                                                                  version=version,
                                                                                  name=name)

    def set_pending(self, build, url, description):
        raise NotImplementedError()

    @gen.coroutine
    def set_success(self, build, url, description):
        raise NotImplementedError()

    @gen.coroutine
    def set_failed(self, build, url, description):
        raise NotImplementedError()

    @gen.coroutine
    def create_release(self, commit_id, tag, version, branch, build_type):
        raise NotImplementedError()

    @gen.coroutine
    def upload_release_file(self, release, board, tag, version, name, fmt, content):
        raise NotImplementedError()

    @gen.coroutine
    def add_s3_release_link(self, release, board, tag, version, name, fmt, s3_url):
        raise NotImplementedError()


def get_service():
    global _service

    if _service is None:
        logger.debug('creating repo service')
        service_class = _SERVICE_CLASSES[settings.REPO_SERVICE]
        _service = service_class()

    return _service


@gen.coroutine
def _fixed_hour_loop():
    last_run_day = 0
    while True:
        yield gen.sleep(60)

        day = datetime.datetime.now().day
        if day == last_run_day:  # prevents running more than once in a day
            continue

        if datetime.datetime.now().hour != settings.NIGHTLY_FIXED_HOUR:
            continue

        last_run_day = day

        logger.debug('running fixed hour nightly build check')
        _check_run_fixed_hour_task()


def _check_run_fixed_hour_task():
    service = get_service()
    service.schedule_nightly_builds_for_new_commits()


def init():
    from thingosdci.reposervices import github
    from thingosdci.reposervices import gitlab
    from thingosdci.reposervices import bitbucket

    _SERVICE_CLASSES['github'] = github.GitHub
    _SERVICE_CLASSES['gitlab'] = gitlab.GitLab
    _SERVICE_CLASSES['bitbucket'] = bitbucket.BitBucket

    logger.debug('starting web server on port %s', settings.WEB_PORT)

    service_class = _SERVICE_CLASSES[settings.REPO_SERVICE]
    service = get_service()

    application = web.Application([
        ('/{}'.format(settings.REPO_SERVICE), service_class.REQUEST_HANDLER_CLASS, {'service': service}),
        ('/trigger', trigger.TriggerRequestHandler, {'service': service}),
    ])

    application.listen(settings.WEB_PORT)

    if settings.NIGHTLY_FIXED_HOUR is not None:
        ioloop.IOLoop.current().spawn_callback(_fixed_hour_loop)




import logging

from tornado import web


logger = logging.getLogger(__name__)


class TriggerRequestHandler(web.RequestHandler):
    def initialize(self, service):
        self.service = service

    def post(self):
        typ = self.get_argument('type')

        if typ == 'nightly':
            branch = self.get_argument('branch')
            self.schedule_nightly_build(branch)

        elif typ == 'tag':
            tag = self.get_argument('tag')
            self.schedule_tag_build(tag)

        else:
            raise web.HTTPError(400, 'unknown type {}'.format(typ))

    def schedule_nightly_build(self, branch):
        self.service.schedule_nightly_build(commit_id=None, branch=branch)

    def schedule_tag_build(self, tag):
        self.service.handle_new_tag(commit_id=None, tag=tag)




import datetime
import functools
import hashlib
import logging
import os
import time

from tornado import gen
from tornado import ioloop

from thingosdci import dockerctl
from thingosdci import loopdevmanager
from thingosdci import settings
from thingosdci import utils


logger = logging.getLogger(__name__)


STATE_PENDING = 'pending'
STATE_RUNNING = 'running'
STATE_ENDED = 'ended'

TYPE_PR = 'pr'
TYPE_NIGHTLY = 'nightly'
TYPE_TAG = 'tag'
TYPE_CUSTOM = 'custom'

_build_queue = []  # pending builds
_current_builds_by_board = {}  # a single build per board is allowed at a time
_current_build_group = None


class BuildException(Exception):
    pass


class Build:
    def __init__(self, repo_service, group, typ, board,
                 commit_id=None, tag=None, pr_no=None, branch=None, version=None,
                 custom_cmd=None, interactive=False, callback=None):

        self.repo_service = repo_service
        self.group = group
        self.type = typ
        self.board = board

        self.commit_id = commit_id
        self.tag = tag
        self.pr_no = pr_no
        self.branch = branch
        self.version = version
        self.custom_cmd = custom_cmd
        self.interactive = interactive

        self.container = None
        self.exit_code = None
        self.begin_time = None
        self.end_time = None
        self.image_files = None

        try:
            self.loop_dev = loopdevmanager.acquire_loop_dev()

        except loopdevmanager.LoopDevManagerException as e:
            logger.error('failed to acquire loop device for %s: %s', self, e)
            self.loop_dev = None

        self._callback = callback
        self._state_change_callbacks = []

        if group:
            group.add_build(self)

    def __str__(self):
        return 'build {}/{}/{}'.format(self.type, self.get_identifier(), self.board)

    def get_identifier(self):
        if self.type == TYPE_PR:
            return self.pr_no

        elif self.type == TYPE_NIGHTLY:
            return self.branch

        elif self.type == TYPE_TAG:
            return self.tag

        else:  # assuming TYPE_CUSTOM
            return 'cmd' + hashlib.sha1(self.custom_cmd.encode()).hexdigest()[:8]

    def get_key(self):
        return '{}/{}/{}'.format(self.repo_service, self.get_identifier(), self.board)

    def set_begin(self, container):
        if self.begin_time is not None:
            raise BuildException('cannot set begin time of build that has already begun')

        self.begin_time = time.time()
        self.container = container

        if container:
            logger.debug('%s has begun on %s', self, container)
            self.container.add_state_change_callback(self._on_container_state_change)

        self._run_state_change_callbacks()

    def set_end(self, exit_code):
        if self.begin_time is None:
            raise BuildException('cannot set end time of build that has not begun')

        if self.end_time is not None:
            raise BuildException('cannot set end time of build that has already ended')

        if self.loop_dev:
            try:
                loopdevmanager.release_loop_dev(self.loop_dev)

            except loopdevmanager.LoopDevManagerException as e:
                logger.error('failed to release loop device %s of %s: %s', self.loop_dev, self, e)

        self.exit_code = exit_code
        self.end_time = time.time()

        # gather image files
        if not self.custom_cmd and not exit_code:  # regular build
            p = os.path.join(settings.OUTPUT_DIR, self.board, '.image_files')
            if os.path.exists(p):
                with open(p, 'r') as f:
                    image_files = f.readlines()

                # simple image file name
                image_files = [f.strip() for f in image_files]

                # full path to image file
                image_files = [os.path.join(settings.OUTPUT_DIR, self.board, 'images', f) for f in image_files]

                # dictionarize by file format/extension
                image_files_by_fmt = {}
                for fmt in settings.IMAGE_FILE_FORMATS:
                    for f in image_files:
                        if f.endswith(fmt):
                            image_files_by_fmt[fmt] = f

                self.image_files = image_files_by_fmt

        lifetime = int(self.end_time - self.begin_time)
        how = ['successfully', 'with error'][bool(exit_code)]
        logger.debug('%s has ended %s (lifetime=%ss)', self, how, lifetime)

        if exit_code:
            # show last 20 lines of container log
            build_log = dockerctl.get_container_log(self.container.id, last_lines=20)

            delimiter = '*' * 3
            logger.error('build failed:\n\n %s\n\n%s\n %s\n', delimiter, build_log, delimiter)

        self._run_state_change_callbacks()

        if self._callback:
            io_loop = ioloop.IOLoop.current()
            io_loop.spawn_callback(self._callback, self)

        if _current_builds_by_board.get(self.board) is self:
            _current_builds_by_board.pop(self.board)

        else:
            logger.warning('%s was not the current build for board %s', self, self.board)

        logger.debug('%d running builds', len(_current_builds_by_board))

    def get_state(self):
        if self.begin_time is None:
            return STATE_PENDING

        if self.end_time is None:
            return STATE_RUNNING

        return STATE_ENDED

    def add_state_change_callback(self, callback):
        self._state_change_callbacks.append(callback)

    def _run_state_change_callbacks(self):
        io_loop = ioloop.IOLoop.current()
        state = self.get_state()

        for callback in self._state_change_callbacks:
            io_loop.spawn_callback(callback, state)

    def _on_container_state_change(self, state):
        if state == dockerctl.CONTAINER_STATE_EXITED:
            self.set_end(self.container.exit_code)


class BuildGroup:
    def __init__(self):
        self.builds = {}

        self._first_build_begun = False
        self._last_build_ended = False

        self._first_build_begin_callbacks = []
        self._last_build_end_callbacks = []

    def add_build(self, build):
        if build.board in self.builds:
            raise BuildException('board already present in build group')

        self.builds[build.board] = build
        build.add_state_change_callback(functools.partial(self._on_build_state_change, build))

    def get_completed_builds(self):
        return [build for build in self.builds.values() if build.get_state() == STATE_ENDED]

    def get_remaining_builds(self):
        return [build for build in self.builds.values() if build.get_state() in [STATE_RUNNING, STATE_PENDING]]

    def get_failed_builds(self):
        return [build for build in self.builds.values() if build.get_state() == STATE_ENDED and build.exit_code]

    def add_first_build_begin_callback(self, callback):
        self._first_build_begin_callbacks.append(callback)

    def add_last_build_end_callback(self, callback):
        self._last_build_end_callbacks.append(callback)

    def _handle_build_begin(self, build):
        if not self._first_build_begun:
            self._first_build_begun = True

            io_loop = ioloop.IOLoop.current()
            for handler in self._first_build_begin_callbacks:
                io_loop.spawn_callback(handler, build)

    def _handle_build_end(self, build):
        if not self.get_remaining_builds():
            if not self._last_build_ended:
                self._last_build_ended = True

                io_loop = ioloop.IOLoop.current()
                for handler in self._last_build_end_callbacks:
                    io_loop.spawn_callback(handler, build)

    def _on_build_state_change(self, build, state):
        if state == STATE_RUNNING:
            self._handle_build_begin(build)

        elif state == STATE_ENDED:
            self._handle_build_end(build)


def schedule_pr_build(repo_service, group, board, commit_id, pr_no):
    return _schedule_build(repo_service, group, TYPE_PR, board,
                           commit_id=commit_id, pr_no=pr_no)


def schedule_nightly_build(repo_service, group, board, commit_id, branch):
    version = utils.branches_format(settings.NIGHTLY_VERSION, branch, datetime.date.today())

    return _schedule_build(repo_service, group, TYPE_NIGHTLY, board,
                           commit_id=commit_id, branch=branch, version=version)


def schedule_tag_build(repo_service, group, board, commit_id, tag, version):
    return _schedule_build(repo_service, group, TYPE_TAG, board,
                           commit_id=commit_id, tag=tag, version=version)


@gen.coroutine
def run_custom_cmd(repo_service, custom_cmd, interactive=False, board='dummyboard'):
    future = gen.Future()

    _schedule_build(repo_service, None, TYPE_CUSTOM, board, custom_cmd=custom_cmd,
                    interactive=interactive, callback=future.set_result)

    build = yield future
    if build.exit_code:
        raise BuildException('custom build command failed')

    return build


def _schedule_build(repo_service, group, typ, board,
                    commit_id=None, tag=None, pr_no=None, branch=None, version=None,
                    custom_cmd=None, interactive=False, callback=None):

    build = Build(repo_service, group, typ, board,
                  commit_id=commit_id, tag=tag, pr_no=pr_no, branch=branch, version=version,
                  custom_cmd=custom_cmd, interactive=interactive, callback=callback)

    logger.debug('scheduling %s', build)

    # if a build with same key is pending, replace it
    index = -1
    for i, b in enumerate(_build_queue):
        if build.get_key() == b.get_key():
            logger.debug('replacing pending %s', b)
            index = i
            break

    if index >= 0:
        _build_queue.remove(_build_queue[index])

    _build_queue.append(build)

    logger.debug('%d queued builds', len(_build_queue))

    return build


@gen.coroutine
def _run_loop():
    global _current_build_group

    while True:
        yield gen.sleep(1)

        if not _build_queue:  # empty queue
            continue

        # wait for a free slot
        if len(_current_builds_by_board) >= settings.DOCKER_MAX_PARALLEL:
            continue

        # clear build group if no build is running
        if len(_current_builds_by_board) == 0:
            _current_build_group = None

        # treat the case where all queued builds correspond to currently building boards
        queued_boards = [b.board for b in _build_queue]
        if all((b in _current_builds_by_board for b in queued_boards)):
            logger.debug('all queued builds correspond to currently building boards, retrying later')
            yield gen.sleep(60)
            continue

        # treat the case where all queued builds correspond to another build group
        queued_groups = set((b.group for b in _build_queue))
        if _current_build_group and all((g is not _current_build_group for g in queued_groups)):
            logger.debug('all queued builds correspond to another build group, retrying later')
            yield gen.sleep(60)
            continue

        build = _build_queue.pop(0)

        logger.debug('dequeued %s (%d remaining queued builds)', build, len(_build_queue))

        if build.board in _current_builds_by_board:
            logger.debug('another build for board %s is currently running, pushing %s back', build.board, build)
            _build_queue.append(build)
            continue

        if _current_build_group and build.group is not _current_build_group:
            logger.debug('%s belongs to another build group, pushing back', build)
            _build_queue.append(build)
            continue

        _current_builds_by_board[build.board] = build
        _current_build_group = build.group

        logger.debug('starting %s (%d running builds)', build, len(_current_builds_by_board))

        # workaround for when running docker over ssh
        custom_cmd = build.custom_cmd or ''
        if not settings.DOCKER_COMMAND.startswith('docker'):
            custom_cmd = '"' + custom_cmd + '"'

        clone_args = ''
        if settings.GIT_CLONE_DEPTH > 0:
            clone_args += '--no-single-branch --depth {}'.format(settings.GIT_CLONE_DEPTH)

        env = {
            'TB_REPO': settings.GIT_URL,
            'TB_GIT_CLONE_ARGS': clone_args,
            'TB_BOARD': build.board,
            'TB_COMMIT': build.commit_id or '',
            'TB_TAG': build.tag or '',
            'TB_PR': build.pr_no or '',
            'TB_BRANCH': build.branch or '',
            'TB_VERSION': build.version or '',
            'TB_CUSTOM_CMD': custom_cmd,
            'TB_CLEAN_TARGET_ONLY': str(settings.CLEAN_TARGET_ONLY).lower(),
            'TB_LOOP_DEV': build.loop_dev or ''
        }

        vol = {
            settings.DL_DIR: '/mnt/dl',
            settings.CCACHE_DIR: '/mnt/ccache',
            settings.OUTPUT_DIR: '/mnt/output'
        }

        try:
            container = dockerctl.run_container(env, vol, build.interactive)

        except dockerctl.DockerException as e:
            logger.error('failed to start build: %s', e, exc_info=True)
            continue

        build.set_begin(container)
        if not container:  # container is None when running an interactive command
            build.set_end(0)


def init():
    io_loop = ioloop.IOLoop.current()
    io_loop.spawn_callback(_run_loop)




import mimetypes


def branches_format(s, branch, now=None):
    s = s.format(branch=branch, Branch=branch.title(), BRANCH=branch.upper())
    if now:
        s = now.strftime(s)

    return s


def encode_multipart_formdata(fields=None, files=None):
    boundary = '----multi-part-form-data-boundary----'
    lines = []

    fields = fields or {}
    files = files or {}

    for key, value in fields.items():
        lines.append('--' + boundary)
        lines.append('Content-Disposition: form-data; name="%s"' % key)
        lines.append('')
        lines.append(value)

    for key, (filename, content) in files.items():
        content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        lines.append('--' + boundary)
        lines.append('Content-Disposition: form-data; name="%s"; filename="%s"' % (key, filename))
        lines.append('Content-Type: %s' % content_type)
        lines.append('')
        lines.append(content)

    lines.append('--' + boundary + '--')
    lines.append('')

    # transform all lines into bytes
    lines = [l.encode() if isinstance(l, str) else l for l in lines]

    body = b'\r\n'.join(lines)
    content_type = 'multipart/form-data; boundary=%s' % boundary

    return content_type, body





WEB_PORT = 4567
WEB_SECRET = 'deadbeef'
WEB_BASE_URL = 'http://www.example.com/thingosdci'

LOG_LEVEL = 'DEBUG'

GIT_URL = 'git@github.com:owner/project.git'
GIT_CLONE_DEPTH = -1
REPO = 'owner/project'
BOARDS = ('raspberrypi', 'raspberrypi2')
IMAGE_FILE_FORMATS = ('.gz', '.xz')

NIGHTLY_BRANCHES = ('master', 'dev')
NIGHTLY_TAG = 'nightly-{branch}'
NIGHTLY_NAME = 'Nightly {Branch}'
NIGHTLY_VERSION = '{branch}%Y%m%d'
NIGHTLY_FIXED_HOUR = None
RELEASE_TAG_REGEX = r'\d{8}'
PULL_REQUESTS = False
CLEAN_TARGET_ONLY = False

DL_DIR = '/var/lib/thingosdci/dl'
CCACHE_DIR = '/var/lib/thingosdci/ccache'
OUTPUT_DIR = '/var/lib/thingosdci/output'
BUILD_LOGS_DIR = '/var/lib/thingosdci/logs'
PERSIST_DIR = '/var/lib/thingosdci/persist'

UPLOAD_REQUEST_TIMEOUT = 600  # seconds
UPLOAD_SERVICE_BUILD_TYPES = ('nightly', 'tag')

REPO_SERVICE = 'github'

GITHUB_ACCESS_TOKEN = 'deadbeef'
GITHUB_REQUEST_TIMEOUT = 20  # seconds

BITBUCKET_USERNAME = 'owner'
BITBUCKET_PASSWORD = 'deadbeef'
BITBUCKET_REQUEST_TIMEOUT = 20  # seconds

GITLAB_PROJECT_ID = 1234
GITLAB_REQUEST_TIMEOUT = 20  # seconds
GITLAB_ACCESS_TOKEN = 'deadbeef'

DOCKER_MAX_PARALLEL = 4
DOCKER_CONTAINER_MAX_AGE = 3600 * 12  # seconds
DOCKER_LOGS_MAX_AGE = 86400 * 31  # seconds
DOCKER_IMAGE_NAME = 'ccrisan/thingos-builder'
DOCKER_COMMAND = 'docker'
DOCKER_COPY_SSH_PRIVATE_KEY = False
DOCKER_ENV_FILE = None

LOOP_DEV_RANGE = (10, 19)  # from /dev/loop10 to /dev/loop19

S3_UPLOAD_BUILD_TYPES = ()  # 'nightly', 'tag'
S3_UPLOAD_ACCESS_KEY = 'deadbeef'
S3_UPLOAD_SECRET_KEY = 'secret'
S3_UPLOAD_BUCKET = 'bucket'
S3_UPLOAD_PATH = 'thingos'
S3_UPLOAD_ADD_RELEASE_LINK = False


try:
    from settingslocal import *

except ImportError:
    pass




import json
import logging
import os

from thingosdci import settings


logger = logging.getLogger(__name__)


def load(name, default=None):
    logger.debug('loading %s', name)

    file_path = os.path.join(settings.PERSIST_DIR, '{}.json'.format(name))
    if os.path.exists(file_path):
        try:
            with open(file_path, 'r') as f:
                return json.load(f)

        except Exception:
            logger.error('cannot read file %s', file_path, exc_info=True)
            raise

    return default


def save(name, value):
    logger.debug('saving %s', name)

    file_path = os.path.join(settings.PERSIST_DIR, '{}.json'.format(name))
    try:
        with open(file_path, 'w') as f:
            json.dump(value, f)

    except Exception as e:
        logger.error('cannot save to file %s', file_path, exc_info=True)
        raise




from setuptools import setup

import thingosdci


name = 'thingosdci'
version = thingosdci.VERSION
description = 'thingOS Docker CI'

setup(
    name=name,
    version=version,

    description=description,
    long_description=description,

    url='https://github.com/ccrisan/thingosdci',

    author='Calin Crisan',
    author_email='ccrisan@gmail.com',

    license='GPLv3',

    packages=['thingosdci', 'thingosdci.reposervices'],

    install_requires=[
        'tornado>=4',
        'redis>=2.10',
        'uritemplate>=3'
    ],

    data_files=[
        ('bin', ['thingosdci.sh'])
    ],

    entry_points={
        'console_scripts': [
            'thingosdci=thingosdci.main:main',
        ],
    }
)




import hashlib
import hmac
import json
import logging
import mimetypes
import uritemplate

from tornado import gen
from tornado import web
from tornado import httpclient

from thingosdci import building
from thingosdci import reposervices
from thingosdci import settings


logger = logging.getLogger(__name__)


_STATUS_CONTEXT = 'thingOS Docker CI'


class GitHubRequestHandler(reposervices.RepoServiceRequestHandler):
    def post(self):
        # verify signature
        remote_signature = self.request.headers.get('X-Hub-Signature')
        if not remote_signature:
            logger.warning('missing signature header')
            raise web.HTTPError(401)

        remote_signature = remote_signature[5:]  # skip "sha1="

        local_signature = hmac.new(settings.WEB_SECRET.encode('utf8'),
                                   msg=self.request.body, digestmod=hashlib.sha1)
        local_signature = local_signature.hexdigest()

        if not hmac.compare_digest(local_signature, remote_signature):
            logger.warning('mismatching signature')
            raise web.HTTPError(401)

        data = json.loads(str(self.request.body.decode('utf8')))
        github_event = self.request.headers['X-GitHub-Event']

        if github_event == 'pull_request':
            action = data['action']
            pull_request = data['pull_request']

            src_repo = pull_request['head']['repo']['full_name']
            dst_repo = pull_request['base']['repo']['full_name']

            commit_id = pull_request['head']['sha']
            pr_no = pull_request['number']

            if action == 'opened':
                self.service.handle_pull_request_open(commit_id, src_repo, dst_repo, pr_no)

            elif action in ['synchronize', 'edited']:
                self.service.handle_pull_request_update(commit_id, src_repo, dst_repo, pr_no)

        elif github_event == 'push':
            if data['head_commit']:
                commit_id = data['head_commit']['id']
                branch_or_tag = data['ref'].split('/')[-1]

                if data['ref'].startswith('refs/tags/'):
                    self.service.handle_new_tag(commit_id, branch_or_tag)

                else:
                    self.service.handle_commit(commit_id, branch_or_tag)


class GitHub(reposervices.RepoService):
    REQUEST_HANDLER_CLASS = GitHubRequestHandler

    def __str__(self):
        return 'github'

    @gen.coroutine
    def _api_request(self, path, method='GET', body=None, extra_headers=None, timeout=settings.GITHUB_REQUEST_TIMEOUT):
        client = httpclient.AsyncHTTPClient()

        access_token = settings.GITHUB_ACCESS_TOKEN
        url = path
        if not (url.startswith('http://') or url.startswith('https://')):
            url = 'https://api.github.com' + url

        if '?' in url:
            url += '&'

        else:
            url += '?'

        url += 'access_token=' + access_token

        headers = {
            'Content-Type': 'application/json',
            'User-Agent': settings.REPO
        }

        headers.update(extra_headers or {})

        if body is not None and not isinstance(body, str) and headers['Content-Type'] == 'application/json':
            body = json.dumps(body)

        response = yield client.fetch(url, headers=headers, method=method, body=body,
                                      connect_timeout=timeout, request_timeout=timeout)
        if not response.body:
            return None

        return json.loads(response.body.decode('utf8'))

    @staticmethod
    def _api_error_message(e):
        if hasattr(e, 'response'):
            try:
                return json.loads(e.response.body.decode('utf8'))

            except Exception:
                return str(e.response.body or e)

        else:
            return str(e)

    @gen.coroutine
    def _set_status(self, commit_id, status, target_url, description, context):
        if not commit_id:
            return

        description = description[:140]  # maximum allowed by github
        path = '/repos/{}/statuses/{}'.format(settings.REPO, commit_id)
        body = {
            'state': status,
            'target_url': target_url,
            'description': description,
            'context': context
        }

        try:
            yield self._api_request(path, method='POST', body=body)

        except Exception as e:
            logger.error('set status failed: %s', self._api_error_message(e))

    @gen.coroutine
    def set_pending(self, build, url, description):
        yield self._set_status(build.commit_id,
                               status='pending',
                               target_url=url,
                               description=description,
                               context=_STATUS_CONTEXT)

    @gen.coroutine
    def set_success(self, build, url, description):
        yield self._set_status(build.commit_id,
                               status='success',
                               target_url=url,
                               description=description,
                               context=_STATUS_CONTEXT)

    @gen.coroutine
    def set_failed(self, build, url, description):
        yield self._set_status(build.commit_id,
                               status='failure',
                               target_url=url,
                               description=description,
                               context=_STATUS_CONTEXT)

    @gen.coroutine
    def create_release(self, commit_id, tag, version, branch, build_type):
        path = '/repos/{}/releases/tags/{}'.format(settings.REPO, tag)

        logger.debug('looking for release %s', version)

        try:
            response = yield self._api_request(path)
            release_id = response['id']
            logger.debug('release %s found with id %s', version, release_id)

        except httpclient.HTTPError as e:
            if e.code == 404:  # no such release, we have to create it
                logger.debug('release %s not present', version)
                release_id = None

            else:
                logger.error('release %s failed: %s', version, self._api_error_message(e))
                return

        except Exception as e:
            logger.error('release %s failed: %s', version, self._api_error_message(e))
            return

        if release_id:
            logger.debug('removing previous release %s', version)

            path = '/repos/{}/releases/{}'.format(settings.REPO, release_id)

            try:
                yield self._api_request(path, method='DELETE')
                logger.debug('previous release %s removed', version)

            except httpclient.HTTPError as e:
                logger.error('failed to remove previous release %s: %s', version, self._api_error_message(e))
                raise

            logger.debug('removing git tag %s', tag)

            custom_cmd = 'git push --delete origin {}'.format(tag)

            try:
                yield building.run_custom_cmd(self, custom_cmd)
                logger.debug('git tag %s removed', tag)

            except Exception as e:
                logger.warning('failed to remove git tag %s: %s', tag, e, exc_info=True)

        logger.debug('creating release %s', version)

        path = '/repos/{}/releases'.format(settings.REPO)
        body = {
            'tag_name': tag,
            'name': version,
            'prerelease': True,
            'draft': build_type == building.TYPE_TAG  # never automatically release a tag build
        }

        if commit_id or branch:
            body['target_commitish'] = commit_id or branch

        try:
            response = yield self._api_request(path, method='POST', body=body)
            release_id = response['id']
            logger.debug('release %s created with id %s', version, release_id)

        except httpclient.HTTPError as e:
            logger.error('release %s failed: %s', version, self._api_error_message(e))
            raise

        return response

    @gen.coroutine
    def upload_release_file(self, release, board, tag, version, name, fmt, content):
        upload_url = release['upload_url']
        ut = uritemplate.URITemplate(upload_url)
        path = ut.expand(name=name)
        content_type = mimetypes.types_map.get(fmt, 'application/octet-stream')

        try:
            yield self._api_request(path, method='POST', body=content, extra_headers={'Content-Type': content_type},
                                    timeout=settings.UPLOAD_REQUEST_TIMEOUT)

        except httpclient.HTTPError as e:
            logger.error('failed to upload file %s: %s', name, self._api_error_message(e))
            return

    @gen.coroutine
    def add_s3_release_link(self, release, board, tag, version, name, fmt, s3_url):
        link = '[{}]({})'.format(name, s3_url)
        release['body'] = (release['body'] or '') + '\n' + link

        path = '/repos/{}/releases/{}'.format(settings.REPO, release['id'])
        logger.debug('updating release %s', version)

        release_body = {
            'body': release['body']
        }
        if tag:
            release_body['tag_name'] = tag

        try:
            yield self._api_request(path, method='PATCH', body=release_body)
            logger.debug('release %s updated with S3 URL %s', version, s3_url)

        except Exception as e:
            logger.error('release %s update failed: %s', version, self._api_error_message(e))




import functools
import logging
import mimetypes
import os
import sys

from tornado import ioloop
from tornado import gen

from thingosdci import building
from thingosdci import dockerctl
from thingosdci import loopdevmanager
from thingosdci import reposervices
from thingosdci import settings
from thingosdci import VERSION


logger = logging.getLogger('thingosdci')


def configure_logging():
    logging.basicConfig(filename=None, level=settings.LOG_LEVEL,
                        format='%(asctime)s: %(levelname)7s: [%(name)s]: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')


def create_dirs():
    os.makedirs(settings.DL_DIR, exist_ok=True)
    os.makedirs(settings.BUILD_LOGS_DIR, exist_ok=True)
    os.makedirs(settings.CCACHE_DIR, exist_ok=True)
    os.makedirs(settings.OUTPUT_DIR, exist_ok=True)
    os.makedirs(settings.PERSIST_DIR, exist_ok=True)


@gen.coroutine
def shell(board=None):
    if board:
        yield building.run_custom_cmd(repo_service=None, custom_cmd='/bin/bash', interactive=True, board=board)

    else:
        yield building.run_custom_cmd(repo_service=None, custom_cmd='/bin/bash', interactive=True)

    ioloop.IOLoop.current().stop()


def main():
    configure_logging()
    logger.info('hello!')
    logger.info('this is thingOS Docker CI %s', VERSION)
    sl = sys.modules.get('settingslocal')
    if sl:
        logger.info('using settings from %s', sl.__file__)

    else:
        logger.warning('using default settings')

    create_dirs()

    mimetypes.init()
    building.init()
    dockerctl.init()
    loopdevmanager.init()

    cmd = len(sys.argv) > 1 and sys.argv[1]
    if cmd == 'shell':
        board = None
        if len(sys.argv) > 2:
            board = sys.argv[2]

        io_loop = ioloop.IOLoop.current()
        io_loop.run_sync(functools.partial(shell, board=board))

    else:
        reposervices.init()
        ioloop.IOLoop.current().start()

    logger.info('bye!')


if __name__ == '__main__':
    main()




import json
import logging

from tornado import gen
from tornado import httpclient

from thingosdci import reposervices
from thingosdci import settings
from thingosdci import utils


logger = logging.getLogger(__name__)


_BUILD_NAME = 'thingOS Docker CI'


class BitBucketRequestHandler(reposervices.RepoServiceRequestHandler):
    def post(self):
        data = json.loads(str(self.request.body.decode('utf8')))
        event = self.request.headers['X-Event-Key']

        if event == 'repo:push':
            changes = data['push']['changes']
            for change in changes:
                if 'new' not in change:
                    continue

                change_type = change['new']['type']
                name = change['new']['name']
                commit_id = change['new']['target']['hash']

                if change_type == 'tag':
                    self.service.handle_new_tag(commit_id, name)

                elif change_type == 'branch':
                    self.service.handle_commit(commit_id, name)

        elif event in ('pullrequest:created', 'pullrequest:updated'):
            pull_request = data['pullrequest']
            commit_id = pull_request['source']['commit']['hash']
            src_repo = pull_request['source']['repository']['full_name']
            dst_repo = pull_request['destination']['repository']['full_name']
            pr_no = pull_request['id']

            if event.endswith('created'):
                self.service.handle_pull_request_open(commit_id, src_repo, dst_repo, pr_no)

            else:  # assuming updated
                self.service.handle_pull_request_update(commit_id, src_repo, dst_repo, pr_no)


class BitBucket(reposervices.RepoService):
    REQUEST_HANDLER_CLASS = BitBucketRequestHandler

    def __str__(self):
        return 'bitbucket'

    @gen.coroutine
    def _api_request(self, path, method='GET', body=None, extra_headers=None,
                     timeout=settings.BITBUCKET_REQUEST_TIMEOUT):

        client = httpclient.AsyncHTTPClient()

        url = path
        if not (url.startswith('http://') or url.startswith('https://')):
            url = 'https://api.bitbucket.org/2.0' + url

        headers = {
            'Content-Type': 'application/json',
            'User-Agent': settings.REPO
        }

        headers.update(extra_headers or {})

        if body is not None and headers['Content-Type'] == 'application/json':
            body = json.dumps(body)

        response = yield client.fetch(url, headers=headers, method=method, body=body,
                                      auth_username=settings.BITBUCKET_USERNAME,
                                      auth_password=settings.BITBUCKET_PASSWORD,
                                      connect_timeout=timeout, request_timeout=timeout)
        if not response.body:
            return None

        return json.loads(response.body.decode('utf8'))

    @staticmethod
    def _api_error_message(e):
        if hasattr(e, 'response'):
            try:
                return json.loads(e.response.body.decode('utf8'))['error']['message']

            except Exception:
                return str(e.response.body)

        else:
            return str(e)

    @gen.coroutine
    def _set_status(self, commit_id, status, url, description, name):
        if not commit_id:
            return

        path = '/repositories/{}/commit/{}/statuses/build'.format(settings.REPO, commit_id)
        body = {
            'state': status,
            'url': url,
            'description': description,
            'name': name,
            'key': commit_id
        }

        try:
            yield self._api_request(path, method='POST', body=body)

        except Exception as e:
            logger.error('set status failed: %s', self._api_error_message(e))
            raise

    @gen.coroutine
    def set_pending(self, build, url, description):
        yield self._set_status(build.commit_id,
                               status='INPROGRESS',
                               url=url,
                               description=description,
                               name=_BUILD_NAME)

    @gen.coroutine
    def set_success(self, build, url, description):
        yield self._set_status(build.commit_id,
                               status='SUCCESSFUL',
                               url=url,
                               description=description,
                               name=_BUILD_NAME)

    @gen.coroutine
    def set_failed(self, build, url, description):
        yield self._set_status(build.commit_id,
                               status='FAILED',
                               url=url,
                               description=description,
                               name=_BUILD_NAME)

    @gen.coroutine
    def create_release(self, commit_id, tag, version, branch, build_type):
        logger.debug('creating tag %s', tag)

        path = '/repositories/{}/refs/tags'.format(settings.REPO)
        body = {
            'name': tag,
            'target': {
                'hash': commit_id
            }
        }

        try:
            yield self._api_request(path, method='POST', body=body)
            logger.debug('tag created')

        except Exception as e:
            msg = self._api_error_message(e)
            if msg.count('already exists'):
                logger.debug('tag already exists')

            else:
                logger.error('tag creation failed: %s', msg)

    @gen.coroutine
    def upload_release_file(self, release, board, tag, version, name, fmt, content):
        path = '/repositories/{}/downloads'.format(settings.REPO)

        content_type, body = utils.encode_multipart_formdata(files={'files': (name, content)})
        headers = {
            'Content-Type': content_type,
            'Content-Length': str(len(body))
        }

        try:
            yield self._api_request(path, method='POST', body=body, extra_headers=headers,
                                    timeout=settings.UPLOAD_REQUEST_TIMEOUT)

        except httpclient.HTTPError as e:
            logger.error('failed to upload file %s: %s', name, self._api_error_message(e))
            return




import json
import logging
import re

from tornado import gen
from tornado import web
from tornado import httpclient

from thingosdci import reposervices
from thingosdci import settings
from thingosdci import utils


logger = logging.getLogger(__name__)


_STATUS_CONTEXT = 'thingOS Docker CI'


class GitLabRequestHandler(reposervices.RepoServiceRequestHandler):
    def post(self):
        # verify signature
        token = self.request.headers.get('X-Gitlab-Token')
        if not token:
            logger.warning('missing token header')
            raise web.HTTPError(401)

        if token != settings.WEB_SECRET:
            logger.warning('mismatching token')
            raise web.HTTPError(401)

        data = json.loads(str(self.request.body.decode('utf8')))
        gitlab_event = self.request.headers['X-Gitlab-Event']

        if gitlab_event == 'Push Hook':
            branch = data['ref'].split('/')[-1]
            for commit in data['commits']:
                commit_id = commit['id']
                self.service.handle_commit(commit_id, branch)

        elif gitlab_event == 'Tag Push Hook':
            tag = data['ref'].split('/')[-1]
            commit_id = data['checkout_sha']
            if commit_id:  # can be null, if tag deleted
                self.service.handle_new_tag(commit_id, tag)

        elif gitlab_event == 'Merge Request Hook':
            attrs = data['object_attributes']
            action = attrs['action']
            commit_id = attrs['last_commit']['id']
            src_repo = attrs['source']['url']
            dst_repo = attrs['target']['url']
            pr_no = attrs['iid']

            if action == 'open':
                self.service.handle_pull_request_open(commit_id, src_repo, dst_repo, pr_no)

            elif action == 'update':
                self.service.handle_pull_request_update(commit_id, src_repo, dst_repo, pr_no)


class GitLab(reposervices.RepoService):
    REQUEST_HANDLER_CLASS = GitLabRequestHandler
    BASE_URL = 'https://gitlab.com'

    def __str__(self):
        return 'gitlab'

    @gen.coroutine
    def _api_request(self, path, method='GET', body=None, extra_headers=None, timeout=settings.GITLAB_REQUEST_TIMEOUT):
        client = httpclient.AsyncHTTPClient()

        url = path
        if not (url.startswith('http://') or url.startswith('https://')):
            url = self.BASE_URL + '/api/v4' + url

        if '?' in url:
            url += '&'

        else:
            url += '?'

        headers = {
            'Content-Type': 'application/json',
            'User-Agent': settings.REPO,
            'Private-Token': settings.GITLAB_ACCESS_TOKEN
        }

        headers.update(extra_headers or {})

        if body is not None and not isinstance(body, str) and headers['Content-Type'] == 'application/json':
            body = json.dumps(body)

        response = yield client.fetch(url, headers=headers, method=method, body=body,
                                      connect_timeout=timeout, request_timeout=timeout)
        if not response.body:
            return None

        return json.loads(response.body.decode('utf8'))

    @staticmethod
    def _api_error_message(e):
        if hasattr(e, 'response'):
            try:
                return json.loads(e.response.body.decode('utf8'))

            except Exception:
                return str(e.response.body or e)

        else:
            return str(e)

    @gen.coroutine
    def _set_status(self, commit_id, state, target_url, description, context):
        if not commit_id:
            return

        path = '/projects/{}/statuses/{}'.format(settings.GITLAB_PROJECT_ID, commit_id)
        body = {
            'state': state,
            'target_url': target_url,
            'description': description,
            'context': context
        }

        try:
            yield self._api_request(path, method='POST', body=body)

        except Exception as e:
            logger.error('set status failed: %s', self._api_error_message(e))

    @gen.coroutine
    def set_pending(self, build, url, description):
        yield self._set_status(build.commit_id,
                               state='pending',
                               target_url=url,
                               description=description,
                               context=_STATUS_CONTEXT)

    @gen.coroutine
    def set_success(self, build, url, description):
        yield self._set_status(build.commit_id,
                               state='success',
                               target_url=url,
                               description=description,
                               context=_STATUS_CONTEXT)

    @gen.coroutine
    def set_failed(self, build, url, description):
        yield self._set_status(build.commit_id,
                               state='failed',
                               target_url=url,
                               description=description,
                               context=_STATUS_CONTEXT)

    @gen.coroutine
    def create_release(self, commit_id, tag, version, branch, build_type):
        logger.debug('looking for tag %s', tag)

        path = '/projects/{}/repository/tags/{}'.format(settings.GITLAB_PROJECT_ID, tag)
        tag_exists = False
        try:
            yield self._api_request(path)
            logger.debug('tag %s found', tag)
            tag_exists = True

        except httpclient.HTTPError as e:
            if e.code == 404:  # no such release, we have to create it
                logger.debug('tag %s not present', tag)

            else:
                logger.error('tag lookup failed: %s', self._api_error_message(e))
                return

        except Exception as e:
            logger.error('tag lookup failed: %s', self._api_error_message(e))
            return

        if not tag_exists:
            logger.debug('creating tag %s', tag)

            path = '/projects/{}/repository/tags'.format(settings.GITLAB_PROJECT_ID)
            try:
                yield self._api_request(path, method='POST', body={'tag_name': tag, 'ref': commit_id or branch})
                logger.debug('tag %s created', tag)

            except Exception as e:
                logger.error('tag creation failed: %s', self._api_error_message(e))
                return

    @gen.coroutine
    def upload_release_file(self, release, board, tag, version, name, fmt, content):
        logger.debug('uploading release file %s', name)

        path = '/projects/{}/uploads'.format(settings.GITLAB_PROJECT_ID)
        content_type, body = utils.encode_multipart_formdata(files={'file': (name, content)})
        headers = {
            'Content-Type': content_type,
            'Content-Length': str(len(body))
        }

        try:
            response = yield self._api_request(path, method='POST', body=body, extra_headers=headers,
                                               timeout=settings.UPLOAD_REQUEST_TIMEOUT)

        except httpclient.HTTPError as e:
            logger.error('failed to upload file %s: %s', name, self._api_error_message(e))
            return

        logger.debug('creating release %s', version)

        markdown = response['markdown']
        m = re.search(r'\[(.*)\]\((.*)\)', markdown)
        if m:
            # make link absolute
            url = self.BASE_URL + '/' + settings.REPO + m.group(2)
            link = '[{}]({})'.format(m.group(1), url)

        else:
            link = markdown

        release_exists = False
        path = '/projects/{}/repository/tags/{}'.format(settings.GITLAB_PROJECT_ID, tag)
        try:
            response = yield self._api_request(path)
            release = response.get('release')
            if release:
                release_exists = True

                # append to existing description
                description = (release.get('description') or '') + '\n\n' + link

            else:
                description = link

        except Exception as e:
            logger.error('release %s failed: %s', version, self._api_error_message(e))
            return

        path += '/release'

        if release_exists:
            try:
                yield self._api_request(path, method='PUT', body={'description': description})
                logger.debug('release %s updated', version)

            except Exception as e:
                logger.error('release %s failed: %s', version, self._api_error_message(e))
                return

        else:
            try:
                yield self._api_request(path, method='POST', body={'description': description})
                logger.debug('release %s created', version)

            except Exception as e:
                logger.error('release %s failed: %s', version, self._api_error_message(e))
                return




import binascii
import hashlib
import hmac
import logging
import mimetypes
import time

from calendar import timegm
from datetime import datetime
from email.utils import formatdate
from urllib.parse import urlparse, quote

from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPError

from thingosdci import settings


AWS_S3_BUCKET_URL = 'http://{bucket}.s3.amazonaws.com/{path}'

logger = logging.getLogger(__name__)


class S3Client(object):
    def __init__(self, access_key, secret_key, bucket, region='us-east-1'):
        self._access_key = access_key
        self._secret_key = secret_key
        self.bucket = bucket
        self._region = region
        self._service = 's3'
        self._request_scope = 'aws4_request'

    def aws_quote(self, text_bytes):
        return quote(text_bytes, safe="").replace('%7E', '~')

    def generate_url(self, path):
        """ Generates a URL for the given file path. """
        return AWS_S3_BUCKET_URL.format(bucket=self.bucket, path=path)

    def _guess_mimetype(self, filename, default='application/octet-stream'):
        if '.' not in filename:
            return default

        prefix, extension = filename.lower().rsplit(".", 1)

        if extension == 'jpg':
            extension = "jpeg"

        return mimetypes.guess_type(prefix + "." + extension)[0] or default

    def _rfc822_datetime(self, t=None):
        if t is None:
            t = datetime.utcnow()

        return formatdate(timegm(t.timetuple()), usegmt=True)

    def _get_credential_scope(self, request_date):
        return '{key}/{date}/{region}/{service}/{scope}'.format(
                key=self._access_key, date=request_date,
                region=self._region, service=self._service,
                scope=self._request_scope)

    def get_canonical_string(self, request_date, host, endpoint, params, headers, method, payload=''):
        request_date_simple = request_date[:8]

        params['AWSAccessKeyId'] = self._access_key
        params['Timestamp'] = request_date
        params['X-Amz-Credential'] = self._get_credential_scope(request_date_simple)
        params['X-Amz-Algorithm'] = params['SignatureMethod']
        params['X-Amz-Date'] = request_date
        params['X-Amz-Expires'] = '86400'

        # create canonical headers
        lowered_headers = {key.lower(): value.strip() for key, value in headers.items()}
        canonicalized_headers = [key + ':' + lowered_headers[key] for key in sorted(lowered_headers.keys())]
        canonicalized_headers = '\n'.join(canonicalized_headers)

        # create signed headers
        canonicalized_signed_headers = [key for key in sorted(lowered_headers.keys())]
        canonicalized_signed_headers = ';'.join(canonicalized_signed_headers)

        params['X-Amz-SignedHeaders'] = canonicalized_signed_headers

        # create canonical query
        canonicalized_query = [
            self.aws_quote(param) + '=' + self.aws_quote(params[param])
            for param in sorted(params.keys())
        ]
        canonicalized_query = '&'.join(canonicalized_query)

        payload_hasher = hashlib.sha256()
        payload_hasher.update(payload)
        payload = binascii.hexlify(payload_hasher.digest())

        canonical_request = (
            method + '\n' +
            endpoint + '\n' +
            canonicalized_query + '\n' +
            canonicalized_headers + '\n\n' +
            canonicalized_signed_headers + '\n' +
            payload.decode()
        )

        return canonical_request

    def get_string_to_sign(self, algorithm, request_date, canonical_request):
        credential_scope = self._get_credential_scope(request_date[:8])
        credential_scope = credential_scope[credential_scope.find('/') + 1:]

        string_to_sign = [algorithm, request_date, credential_scope]

        hasher = hashlib.sha256()
        hasher.update(canonical_request.encode())
        canonical_request = binascii.hexlify(hasher.digest())

        string_to_sign.append(canonical_request.decode())

        return '\n'.join(string_to_sign)

    def calculate_signature(self, request_date, host, endpoint, params, headers, method, payload=''):
        algorithm = params['SignatureMethod']

        canonical_request = self.get_canonical_string(request_date, host, endpoint, params, headers, method, payload)
        string_to_sign = self.get_string_to_sign(algorithm, request_date, canonical_request)

        request_date_simple = request_date[:8]

        digestmod = hashlib.sha256
        kdate = hmac.new(('AWS4' + self._secret_key).encode(), request_date_simple.encode(), digestmod).digest()
        kregion = hmac.new(kdate, self._region.encode(), digestmod).digest()
        kservice = hmac.new(kregion, self._service.encode(), digestmod).digest()
        ksigning = hmac.new(kservice, self._request_scope.encode(), digestmod).digest()

        signature = hmac.new(ksigning, string_to_sign.encode(), digestmod).digest()

        return binascii.hexlify(signature)

    def sign_request(self, host, endpoint, params, headers, method, payload=''):
        request_date = time.strftime('%Y%m%dT%H%M%SZ', time.gmtime())

        signature = self.calculate_signature(request_date, host, endpoint, params, headers, method, payload)

        canonical_query = [
            self.aws_quote(param) + '=' + self.aws_quote(params[param])
            for param in sorted(params.keys())
        ]
        canonical_query = '&'.join(canonical_query)

        return 'http://{host}{endpoint}?{query}&X-Amz-Signature={signature}'.format(
                host=host, endpoint=endpoint, query=canonical_query, signature=self.aws_quote(signature))

    @gen.coroutine
    def upload(self, path, data, headers=None):
        if isinstance(data, str):
            data = data.encode('utf8')

        headers = headers or {}

        client = AsyncHTTPClient()
        method = 'PUT'
        url = self.generate_url(path)
        url_object = urlparse(url)
        params = {
            'SignatureMethod': 'AWS4-HMAC-SHA256'
        }

        headers.update({
            'Content-Length': str(len(data)),
            'Content-Type': self._guess_mimetype(path),
            'Date': self._rfc822_datetime(),
            'Host': url_object.hostname,
            'X-Amz-Content-sha256': hashlib.sha256(data).hexdigest(),
        })

        try:
            response = yield client.fetch(
                self.sign_request(
                    url_object.hostname,
                    url_object.path,
                    params,
                    headers,
                    method,
                    data
                ),
                method=method,
                body=data,
                connect_timeout=settings.UPLOAD_REQUEST_TIMEOUT,
                request_timeout=settings.UPLOAD_REQUEST_TIMEOUT,
                headers=headers
            )

        except HTTPError as error:
            logger.error(error)
            if error.response:
                logger.error(error.response.body)

            raise gen.Return(None)

        raise gen.Return(response)
